@@ -28,10 +28,12 @@ PIPELINE_TYPE = "pipeline_type"
 STATE = "state"
 VERSION = "version"
 PY_VERSION = "py_version"
-PACKAGES = "packages"
+REQUIREMEMTS = "requirements"
+CONDA_ENV = "conda_env"
 VALID_VAEX_HEADERS = ["*.csv", "*.hdf5", "*.parquet", "*.arrow"]
 DEFAULT_SUFFIX = ".parquet"
 BYTE_DELIMITER = b'###'
+VENV = 'venv'
 
 # App
 CORS_ORIGINS = os.getenv('ORIGINS', '*').split(',')
@@ -39,3 +41,6 @@ ALLOW_METHODS = os.getenv('ALLOW_METHODS', '*').split(',')
 ALLOW_HEADERS = os.getenv('ALLOW_HEADERS', '*').split(',')
 ALLOW_CREDENTIALS = bool(strtobool(os.getenv('ALLOW_CREDENTIALS', 'True')))
 ALLOW_CORS = bool(strtobool(os.getenv('ALLOW_CORS', 'True')))
+
+# package
+GOLDILOX_VERSION = "0.0.10"
